use osu_file_parser::{
    hitobjects::{HitObject, HitObjectParams::*, SlideParams, CurveType},
    osu_file::hitobjects,
    Decimal as OsuDecimal,
    OsuFile,
    Position
};
use rust_decimal::{Decimal, prelude::ToPrimitive};
use kurbo::{BezPath, PathEl, Point, QuadSpline};

mod types;
use types::*;

mod config;
use config as cfg;

/*
    Calculates the maximum velocity (mm/ms) given the the start and end
    position, acceleration, and required total move time.

    This uses an equation from 19aksh https://math.stackexchange.com/a/3304761
    rearranged to solve for v.
    
    In loose mathematical format:

        2t - sqrt(4t² - 8d (1 / a - 1 / -a))
    v = -----------------------------------
                2 (1 / a - 1 / -a)

    Where:
        t = time (ms)
        d = distance (mm)
        a = kinematic acceleration (mm/1000² since 1s == 1000ms)
        v = velocity (mm/ms)

    (multiply v by 60_000 to get the feedrate (mm/min))

    There are some combinations of t, d, and a that can result in no solution,
    usually because the acceleration and deceleration periods overlap. In this
    case, the acceleration should be increased until a solution is found. To
    compensate for the possibility of this exceeding the MAX_ACCEL, this will
    normally calculate with half of MAX_ACCEL. If no solution is found even
    at MAX_ACCEL, an error will be thrown.
 */
fn max_velocity(
    p1: PositionF64,
    p2: PositionF64,
    accel: f64,
    ms: f64,
) -> f64 {
    let dist = p1 >> p2;
    let accel_divided = accel / 1_000_000.0; // accel / 1000² since 1s == 1000ms 
    let accel_factor = 1.0 / accel_divided - 1.0 / -accel_divided;
    let sqrt_fac = (4.0 * ms.powf(2.0) - 8.0 * dist * accel_factor).sqrt();
    (2.0 * ms - sqrt_fac) / (2.0 * accel_factor)
}

fn find_bbox_scale_factor(
    box_size: (f64, f64),
    target_size: (f64, f64),
    rotation: f64
) -> f64 {
    let target_size = if rotation == 90.0 || rotation == 270.0 {
        (target_size.1, target_size.0)
    } else {
        target_size
    };
    if target_size.0 > target_size.1 {
        target_size.1 / box_size.1
    } else {
        target_size.0 / box_size.0
    }
}

fn solve_feedrate_and_accel(
    p1: PositionF64,
    p2: PositionF64,
    starting_accel: f64,
    ms: f64
) -> (f64, f64) {
    let mut accel = starting_accel;
    let mut max_v: f64;

    loop {
        max_v = max_velocity(p1, p2, accel, ms);
        if !max_v.is_nan() { break }
        accel += cfg::ACCEL_RETRY_STEP;
        if accel >= cfg::MAX_ACCEL {
            panic!("couldn't calculate max_v for dist {} within {}ms", p1 >> p2, ms);
        }
    }

    let speed = max_v * 1_000.0; // mm/s
    let feedrate = speed * 60.0; // mm/min

    eprintln!(
        ";   feedrate: {:.2} ({:.2}mm/s) ({})->({}) dist {}, accel={}, Δms={}",
        feedrate, speed,
        p1, p2,
        p1 >> p2,
        accel,
        ms
    );

    (feedrate, accel)
}

fn generate_gcode_move(
    destination: PositionF64,
    accel: f64,
    last_accel: f64,
    feedrate: f64
) -> String {
    let mut output = String::new();
    if accel != last_accel {
        output = format!("M204 S{}\n", accel);
    }

    output = format!(
        "{}G0 F{:.4}\nG0 X{} Y{}",
        output,
        feedrate,
        destination.x, destination.y
    );

    println!("{}", output);
    output
}

fn osu_position_to_kurbo_point(pos: &Position) -> Point {
    Point {
        x: pos.x.get().clone().expect_left("why").to_f64().expect("why"),
        y: pos.y.get().clone().expect_left("why").to_f64().expect("why")
    }
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    /*
        # Calculating moves and optimizing

        The way this system works is by iterating through the hitobjects and
        "accumulating" state, then creating *moves* from said state. On each
        iteration, we don't "move to the next point", instead, we decide whether
        or not to generate gcode to move to the last point. This also means that
        every move starts at a point in the past. Here is a pseudocode
        representation:

        start of the move = the last bounding box corner
        last point = the last bounding box corner

        if it's the first point:
            tell the machine to pause
        otherwise:
            if this point is in the same location as the last point:
                enable stack mode
            otherwise:
                if stack mode is on:
                    generate gcode to move to the first point in the stack
                    generate gcode to dwell for the duration of the stack
                    disable stack mode
                otherwise, if the hitobject qualifies for grouping:
                    increment the group size
                otherwise:
                    generate gcode to move to the last point
                store the last point as the start of the move
        store the current point as the last point
        increment total time

        Keep in mind that a line of gcode only tells the pen to move to another
        position, and the starting location is based on its current position,
        which was the destination of the last move.
    */

    let src = cfg::OSU.parse::<OsuFile>()?;
    let binding = src.hitobjects.expect("Beatmap has no hit objects!");

    let scale_fac = find_bbox_scale_factor(
        cfg::BOUNDING_BOX_SIZE,
        cfg::AREA_SIZE,
        cfg::ROTATION
    );

    println!("G90\nG0 Z0");
    let corners = [
        cfg::AREA_ORIGIN,
        (cfg::AREA_ORIGIN.0, cfg::AREA_MAX.1),
        (cfg::AREA_MAX.0, cfg::AREA_MAX.1),
        (cfg::AREA_MAX.0, cfg::AREA_ORIGIN.1)  
    ];

    for corner in corners {
        println!("G0 X{} Y{} F7500", corner.0, corner.1);
    }

    let final_corner = PositionF64::from_tuple(corners[corners.len() - 1]);

    // An extra element is needed at the end of the hitobjects vec
    let temp_binding = binding.0.clone();
    let mut last_hitobject = temp_binding.last().unwrap().clone();
    last_hitobject.position = Position {
        x: OsuDecimal::new(Decimal::from_f64_retain(final_corner.x).unwrap()),
        y: OsuDecimal::new(Decimal::from_f64_retain(final_corner.y).unwrap()),
    };

    let mut hitobjects = binding.0;
    hitobjects.push(last_hitobject.clone());
    let hitobjects = hitobjects.iter();

    let mut s = IteratorState {
        total_time: 0,

        last_point: final_corner,
        last_point_ms: 0,
        last_accel: 0.0,

        move_start: final_corner,
        move_start_ms: 0,

        group_linear_vel: 0.0,
        group_slope: 0.0,
        group_size: 0,

        stack: false,
        stack_start_ms: 0,

        osu_file: src
    };

    for (idx, obj) in hitobjects.enumerate() {
        iteration(idx, obj, scale_fac, &mut s);
    }

    eprintln!("total time should be {}ms", s.total_time);

    Ok(())
}

fn iteration(
    idx: usize,
    obj: &HitObject,
    scale_fac: f64,
    s: &mut IteratorState
) {
    let current_ms = obj.time.get().clone().unwrap_left().to_u32().unwrap();
    let ms_since_last_point = current_ms - s.last_point_ms;

    let current_point = PositionF64
        ::from_position(obj.position.clone())
        .unwrap()
        .convert_to_machine_space(
            scale_fac,
            cfg::BOUNDING_BOX_SIZE,
            cfg::AREA_ORIGIN,
            cfg::AREA_SIZE,
            cfg::ROTATION
        );

    eprintln!("point {}", idx);

    if idx == 0 {
        // Pause right before the first move
        println!("PAUSE");
    } else if let HitCircle = obj.obj_params {
        handle_hitcircle(current_ms, ms_since_last_point, current_point, s);
    } else  {
        if s.stack {
            end_stack(s);
        } else {
            end_move(s);
        }

        match &obj.obj_params {
            Slider(params) => {
                handle_slider(obj.position.clone(), &params, s);
            },
            _ => {}
        }
    }

    s.total_time += ms_since_last_point;
    s.last_point = current_point;
    s.last_point_ms = current_ms;
}

fn handle_hitcircle(
    current_ms: u32,
    ms_since_last_point: u32,
    current_point: PositionF64,
    s: &mut IteratorState
) {
    if s.last_point == current_point && !s.stack {
        // Begin stacking if the last point is the same as the current one
        s.stack = true;
        s.stack_start_ms = s.last_point_ms;
    } else if s.last_point != current_point {
        let linear_vel = (s.last_point >> current_point) / ms_since_last_point as f64;
        let slope = s.last_point / current_point;
        let can_group = s.group_linear_vel < cfg::GROUP_VEL_LENIENCY
            && slope == s.group_slope;

        if s.stack {
            end_stack(s);
        } else if can_group {
            println!("; !! adding point to group");
            s.group_size += 1;
        } else {
            end_move(s);
            s.group_linear_vel = linear_vel;
            s.group_slope = slope;
        }
    }
}

/*
    This essentially works the same way as the base iteration. It creates
    subpaths out of a set of control points by "accumulating" previous points
    and deciding on whether or not to create a new subpath based on the next
    point.

    Sliders are saved as different types if they have multiple types of curves.
    Essentially, a path becomes a bézier if it isn't completely linear. Perfect
    circle subpaths get approximated as béziers, too.
*/
fn handle_slider(
    pos: Position,
    params: &SlideParams,
    s: &mut IteratorState
) {
    if params.curve_points.len() == 0 as usize { return; }

    let points = &params.curve_points
        .iter()
        .map(|p| osu_position_to_kurbo_point(&p.0));

    let ticks = match params.curve_type {
        CurveType::Linear => { },
        CurveType::PerfectCircle => { },
        _ => { }
    }
}

fn calculate_linear_ticks(
    pos: Position,
    points: Vec<Point>,
    
) -> Vec<PositionF64> {
    if points.len() == 0 { panic!("wtf") }

    let mut dist = 0.0;
    let dist_between_ticks = 

    let mut ticks = Vec::<PositionF64>::new();

    let mut dist_remainder = dist;
    let mut point_iter = points.iter();
    let mut p1 = PositionF64::from_point(*point_iter.next().unwrap());

    'each_tick: while {

    }
    loop {
        let next_point = point_iter.next();
        let p2 = PositionF64::from_point(*point_iter.next().unwrap());
        let dist = p1 >> p2;

        if dist_remainder < dist {
            let fac = dist_remainder / dist;

            ticks.push(PositionF64 {
                x: (p2.x - p1.x) * fac,
                y: (p2.y - p1.y) * fac
            });
            break;
        } else if dist_remainder == dist {
            ticks.push(p2);
            break;
        } else {
            dist_remainder -= dist;
            p1 = p2;
        }
    }

    ticks
}

fn calculate_bezier_ticks(
    pos: Position,
    points: Vec<Point>,
    slider_mul: f64,
    slider_ticks: u8
) -> Vec<PositionF64> {
    let last_point: Point;
    let mut segment_vertices = Vec::<Point>::new();
    segment_vertices.push(last_point);

    let mut split_path = BezPath::from_vec(vec![PathEl::MoveTo(last_point)]);

    for point in points.iter() {
        // Previous path ended, new subpath begins
        if *point == last_point {
            match segment_vertices.len() {
                0 => { panic!("bruhhhhhh {:?}", pos) },
                1 => { split_path.line_to(last_point) },
                _ => {
                    QuadSpline
                        ::new(segment_vertices)
                        .to_quads()
                        .for_each(|quad| split_path.quad_to(quad.p1, quad.p2));
                }
            }

            segment_vertices = Vec::<Point>::new();
            segment_vertices.push(last_point);
            segment_vertices.push(*point);
        } else {
            segment_vertices.push(*point);
        }
    }
    vec![PositionF64::new(10.0, 10.0)]
}

fn end_stack(s: &mut IteratorState) {
    // Generates gcode to move to the stack and wait there for the
    // totkal duration of its hitobjects, and resets the stack state.
    let (feedrate, accel) = solve_feedrate_and_accel(
        s.move_start,
        s.last_point,
        cfg::DEFAULT_ACCEL,
        (s.stack_start_ms - s.move_start_ms) as f64
    );

    generate_gcode_move(s.last_point, accel, s.last_accel, feedrate);
    println!("G4 P{}", s.last_point_ms - s.stack_start_ms);

    s.last_accel = accel;

    s.move_start_ms = s.last_point_ms;
    s.move_start = s.last_point;

    s.stack = false;
}

fn end_move(s: &mut IteratorState) {
    // Generates gcode for a move and starts a new one.

    // Use higher acceleration when moving across a group
    let accel = if s.group_size > 0 {
        cfg::GROUP_ACCEL
    } else {
        cfg::DEFAULT_ACCEL
    };

    let (feedrate, accel) = solve_feedrate_and_accel(
        s.move_start,
        s.last_point,
        accel,
        (s.last_point_ms - s.move_start_ms) as f64
    );

    generate_gcode_move(s.last_point, accel, s.last_accel, feedrate);

    s.last_accel = accel;

    s.move_start_ms = s.last_point_ms;
    s.move_start = s.last_point;

    s.group_size = 0;
}

        /*
        println!(
            "x={} y={} Δms={} type={}",
            obj.position.x,
            obj.position.y        let move_p1_to_p1_ms = p1_ms
            .checked_sub(move_p1_ms)
            .unwrap()
            .to_f64()
            .unwrap();,
            delta_ms,
            match obj.obj_params {
                HitCircle => String::from("HitCircle"),
                Slider(s) => format!(
                    "Slider type={} points={} slides={} length={}",
                    match s.curve_type {
                        Bezier => "Bezier",
                        Centripetal => "Centripetal",
                        Linear => "Linear",
                        PerfectCircle => "PerfectCircle",
                        _ => "Other"
                    },
                    s.curve_points.len(),
                    s.slides,
                    s.length.get().clone()
                ),
                Spinner => String::from("Spinner"),
                _ => String::from("Other")
            }
        );
        */